/*  undump *  Ver. 1.1 : 30/7/93 *  Author: D.J.Nixon 
 *
 *  (C.) 1990 University of Liverpool
*/

/*
 *
 *  Create "pre-loaded" programmes. (HP-UX 9.01 s700)
 *
 *  Given an executable and its core image 'undump' will combine 
 *  the executable's header and text with data from the core file
 *  to build an "un-dumped" executable.
 *
*/


#include <stdio.h>  
#include <fcntl.h>
#include <signal.h>  
#include <errno.h>
#include <unistd.h>
#include <a.out.h>
#include <sys/types.h>
#include <sys/param.h>  
#include <sys/stat.h>
#include <sys/core.h>

#define FILE_P 0711   /* undumped executable file permissions */


/* global data */
extern int errno;    
int ex,core,nex;   /* file descriptors */
char *undumped="b.out";  

/* functions */
#ifndef __hpux
long lseek();
#endif
void fatal_err();  
void help();
int handler();  
off_t get_core_data_sz();
off_t copyf();
int check_core_date();
void bell();


void main(argc, argv)
int argc;
char *argv[4];
{
int gflag=0;
off_t bytes,
      orig_unloadable_sp_location,
      core_data_sz;
struct header hdr;     
struct som_exec_auxhdr hpuxhdr;
FILE *logfile=stdout;

 /*
  *  Parse command line arguments.
  */
  {
  register int n=2;
    if (argc < 3) help();
    /* silent by default */
    if ((logfile=fopen("/dev/null", "a")) == NULL) logfile=stdout;
    while (++n < argc)
      if ((argv[n][0]) == '-')
        switch(argv[n][1])
        {
        case 'g':  /* include debugging data */
                   ++gflag;
                   break;
        case 'o':  /* name undumped object file */
                   if (((argc - 1) == n) || (argv[++n][0] == '-')) help();
                   undumped=argv[n];
                   break;
        case 'h':  help();
        case 'v':  /* verbose output mode */
                   logfile=stdout;
                   break;
        default:   fprintf(stderr,"Unknown option: -%c\n",argv[n][1]);
                   exit(1);
        }
      else
        help();
    if ((ex=open(argv[1], O_RDONLY)) == -1) {
      fatal_err("Can't open executable file"); 
    }
    if ((core=open(argv[2], O_RDONLY)) == -1) {
      fatal_err("Can't open core file");
    }
    if ((nex=open(undumped, O_WRONLY | O_CREAT, FILE_P)) == -1) {
      fatal_err("Can't create undumped file");
    }  
  }

  /*
   *  Set signals.
   */
  signal(SIGQUIT, SIG_IGN);
  signal(SIGINT, handler);

  /*
   *  Read executable's headers.
   */
      /* Spectrum Object Module header */
  if ((read(ex, &hdr, sizeof(struct header))) < sizeof(struct header)) 
                           fatal_err("Could not read Spectrum header");
  /* check executable file is for HPPA */    
  if (!_PA_RISC_ID(hdr.system_id)) fatal_err("Bad system magic number");      
  switch (hdr.a_magic)
  {
  case EXEC_MAGIC:
    fputs("Normal executable\n",logfile);
    break;
  case SHARE_MAGIC:
    fputs("Shared executable\n",logfile);
    break;
  case DEMAND_MAGIC:
    fputs("Demand paged executable\n",logfile);
    break;
  default:
    fatal_err("Bad file type magic number");
  }
  /* auxiliary HP-UX header */
  if ((read(ex, &hpuxhdr, sizeof(struct som_exec_auxhdr))) < sizeof(struct som_exec_auxhdr)) 
                           fatal_err("Could not read HP-UX header");
  /* check executable is for HP-UX */
  if (!(hpuxhdr.som_auxhdr.type == HPUX_AUX_ID))
                           fatal_err("Not an HP-UX executable!");
  fprintf(logfile,"Text: %u Data: %u BSS: %u\n",hpuxhdr.exec_tsize,hpuxhdr.exec_dsize,hpuxhdr.exec_bsize);

  /* get size of initialised data, BSS and heap from core */
  if (!check_core_date(argv[1], argv[2]))
                           fatal_err("Core is older than executable");
  if (!(core_data_sz=get_core_data_sz()))
                           fatal_err("Could not get size of core data");
  fprintf(logfile,"Core Data/BSS: %u\n",core_data_sz);


  /*
   *  Adjust undumped executable's HP-UX header.
   *  Assumes that the executable file data comes after the rest.
   */
  hdr.som_length+=core_data_sz - hpuxhdr.exec_dsize;
  hpuxhdr.exec_bsize=0L;
  orig_unloadable_sp_location=hdr.unloadable_sp_location;
  hpuxhdr.exec_dsize=core_data_sz;
  if (gflag)
    hdr.unloadable_sp_location=hpuxhdr.exec_dfile + core_data_sz;
  /* checksum now wrong (what uses the checksum?). Should really use
     HPs checksum routine */
                           
  /*
   *  Build undumped executable: executable header, text, records and symbols,
   *  + core file data object.
   */
     /* headers and text */
  fprintf(logfile,"\nCreating undumped executable:\n");
  if ((write(nex, &hdr, sizeof(struct header))) < sizeof(struct header)) 
                          fatal_err("Error writing Spectrum header"); 
  if ((write(nex, &hpuxhdr, sizeof(struct som_exec_auxhdr))) < sizeof(struct som_exec_auxhdr)) 
                          fatal_err("Error writing HP-UX header"); 
  lseek(ex, hpuxhdr.exec_tfile, SEEK_SET); 
  lseek(nex, hpuxhdr.exec_tfile, SEEK_SET);     
  fprintf(logfile,"Copying %u bytes of executable text from %s\n",hpuxhdr.exec_tsize,argv[1]);  
  if ((bytes=copyf(ex, nex, hpuxhdr.exec_tsize)) < hpuxhdr.exec_tsize ) {
    if (bytes < 0) {
      fatal_err("Error copying executable text");
    } else {
      fatal_err("At EOF in executable text");
    }  
  }
  /* space dictionary */
  lseek(ex, hdr.space_location, SEEK_SET); 
  lseek(nex, hdr.space_location, SEEK_SET);     
  fprintf(logfile,"Copying %u space dictionary records\n",hdr.space_total);  
  if ((bytes=copyf(ex, nex, (hdr.space_total * sizeof(struct space_dictionary_record))) < (hdr.space_total * sizeof(struct space_dictionary_record)))) {
    if (bytes < 0) {
      fatal_err("Error copying space records");
    } else { 
      fatal_err("At EOF in space records");
    }
  } 
  /* subspace dictionary */
  lseek(ex, hdr.subspace_location, SEEK_SET); 
  lseek(nex, hdr.subspace_location, SEEK_SET);     
  fprintf(logfile,"Copying %u subspace dictionary records\n",hdr.subspace_total);  
  if ((bytes=copyf(ex, nex, (hdr.subspace_total * sizeof(struct subspace_dictionary_record))) < (hdr.subspace_total * sizeof(struct subspace_dictionary_record)))) {
    if (bytes < 0) {
      fatal_err("Error copying subspace records");
    } else { 
      fatal_err("At EOF in subspace records");
    }
  } 
  /* space strings */
  lseek(ex, hdr.space_strings_location, SEEK_SET); 
  lseek(nex, hdr.space_strings_location, SEEK_SET);     
  fprintf(logfile,"Copying %u bytes of space strings\n",hdr.space_strings_size);  
  if ((bytes=copyf(ex, nex, hdr.space_strings_size)) < hdr.space_strings_size) {
    if (bytes < 0) {
      fatal_err("Error copying space strings");
    } else { 
      fatal_err("At EOF in space strings");
    }
  } 
  /* compiler records */
  lseek(ex, hdr.compiler_location, SEEK_SET); 
  lseek(nex, hdr.compiler_location, SEEK_SET);     
  fprintf(logfile,"Copying %u compiler records\n",hdr.compiler_total);  
  if ((bytes=copyf(ex, nex, (hdr.compiler_total * sizeof(struct compilation_unit))) < (hdr.compiler_total * sizeof(struct compilation_unit)))) {
    if (bytes < 0) {
      fatal_err("Error copying compiler records");
    } else { 
      fatal_err("At EOF in compiler records");
    }
  } 
  /* symbol table */
  lseek(ex, hdr.symbol_location, SEEK_SET); 
  lseek(nex, hdr.symbol_location, SEEK_SET);     
  fprintf(logfile,"Copying %u symbol records\n",hdr.symbol_total);  
  if ((bytes=copyf(ex, nex, (hdr.symbol_total * sizeof(struct symbol_dictionary_record))) < (hdr.symbol_total * sizeof(struct symbol_dictionary_record)))) {
    if (bytes < 0) {
      fatal_err("Error copying symbol records");
    } else { 
      fatal_err("At EOF in symbol records");
    }
  } 
  /* fixups */
  lseek(ex, hdr.fixup_request_location, SEEK_SET); 
  lseek(nex, hdr.fixup_request_location, SEEK_SET);     
  fprintf(logfile,"Copying %u fixup requests\n",hdr.fixup_request_total);  
  if ((bytes=copyf(ex, nex, (hdr.fixup_request_total * sizeof(struct fixup_request_record))) < (hdr.fixup_request_total * sizeof(struct fixup_request_record)))) {
    if (bytes < 0) {
      fatal_err("Error copying fixup records");
    } else { 
      fatal_err("At EOF in fixup records");
    }
  } 
  /* symbol strings */
  lseek(ex, hdr.symbol_strings_location, SEEK_SET); 
  lseek(nex, hdr.symbol_strings_location, SEEK_SET);     
  fprintf(logfile,"Copying %u bytes of symbol strings\n",hdr.symbol_strings_size);  
  if ((bytes=copyf(ex, nex, hdr.symbol_strings_size)) < hdr.symbol_strings_size) {
    if (bytes < 0) {
      fatal_err("Error copying symbol strings");
    } else { 
      fatal_err("At EOF in symbol strings");
    }
  } 
  /* contiguous data/BSS (from core file) */
  lseek(nex, hpuxhdr.exec_dfile, SEEK_SET); 
  fprintf(logfile,"Undumping %u bytes from core data object ...\n",core_data_sz);  
  if ((bytes=copyf(core, nex, core_data_sz)) < core_data_sz) {
    if (bytes < 0) {
      fatal_err("Error copying core data object");
    } else { 
      fatal_err("At EOF in core");
    }
  } 
  /* debugging data */
  if (gflag)
  {
    if (hdr.unloadable_sp_size == 0)
      fputs("No debugging information\n",logfile);  
    else
    {
      lseek(ex, orig_unloadable_sp_location, SEEK_SET); 
      lseek(nex, hdr.unloadable_sp_location, SEEK_SET); 
      fprintf(logfile,"Copying %u bytes of debug information\n",hdr.unloadable_sp_size);  
      if ((bytes=copyf(ex, nex, hdr.unloadable_sp_size)) < hdr.unloadable_sp_size)    {
        if (bytes < 0)
          fatal_err("Error copying debug data");
        else  
          fprintf(logfile,"... copied  %u bytes of debug information\n",bytes);  
      } 
    }
  }

  /* finis */
  fputs("\n... done!\n\n",logfile);
  bell();
  exit(0);
}

/* 
 *  Copies specified number of bytes from file f1 to file f2.
 *  Returns the number of bytes sucessfully copied.
 */
off_t copyf(f1,f2,bytes)
off_t f1,f2,bytes;
{
char buf[BUFSIZ]; 
register off_t count=0,n,diff;
  while (count < bytes)
  {
    switch (n=read(f1, buf, BUFSIZ))
    {
      case 0:   /* EOF */
        return(count);
      case -1:
        return(-1);
      default:
        count+=n;
        if (count > bytes)  /* adjust */
        {
          diff=(count - bytes);
          n-=(diff);
          count-=(diff);
        }
    }
    if (write(f2, buf, n) == -1) return(-2);                          
  }
  return(count);
}

void fatal_err(message)
char *message;
{
  if (errno) perror(message);
  else
             fprintf(stderr,"%s\n",message);
  close(nex);
  unlink(undumped);   /* tidy up */
  exit(1);
}

void help()
{
  puts("Usage: undump obj1 core [-g] [-o obj2] [-v (verbose)]");
  exit(1);
}

/*
 *  Removes undumped file on interrupt.
 */
handler(sig)
int sig;
{
  puts("\nInterrupt: removing undumped executable");
  close(nex);
  unlink(undumped);
  exit(1);   
}

/*
 *  Return TRUE if the core is younger than 
 *  the executable.
 */
check_core_date(exec_fpath, core_fpath)
char *exec_fpath, *core_fpath;
{
struct stat statbuf;
time_t exec_mtime;
  if (stat(exec_fpath, &statbuf) == -1) fatal_err("Could not stat exec");
  exec_mtime=statbuf.st_mtime;
  if (stat(core_fpath, &statbuf) == -1) fatal_err("Could not stat core");
  if (exec_mtime > statbuf.st_mtime) return(0);
  return(1);
}

/*
 *  Returns the size of 
 *  the core CORE_DATA object.
 */
off_t get_core_data_sz()
{
struct corehead corehdr;
  do
  {
    if ((read(core, &corehdr, sizeof(struct corehead))) < sizeof(struct corehead))
                           fatal_err("Could not read Core header");
    if (corehdr.type != CORE_DATA) 
       if (lseek(core, corehdr.len, SEEK_CUR) == -1)
                           fatal_err("Seek error in Core");
  }
  while (corehdr.type != CORE_DATA);
  return (corehdr.len);
}

/* 
 *  This just beeps on my HP Xterminal
 *  but may sound irritating on dumb terminals.
 */
void bell()
{
#define BELL 0x07
int c=BELL;
register int n,i;
  for (i=0; i < 5; ++i)
  {
    for (n=0; n < 10000000; ++n);
    putchar(c);
  }
}
